unit KnobsConversions;

{

  (C) COPYRIGHT 2014 .. 2015 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Classes, System.Math, Vcl.Graphics,

  ComplexMath, KnobsUtils, Rationals;


const

  TWO_PI  = PI * 2;                                  // Two times Pi, for angle conversions
  HALF_PI = PI / 2;                                  // Pi / 2      ; for fade tables
  DELTA   = 1e-6;                                    // Small value for equality comparisons

  KIBI    = 1024;                                    // Binary multiplication factors
  MEBI    = KIBI * KIBI;
  GIBI    = KIBI * MEBI;

  NO_NOTE = 1.23456e7;                               // Magic number for sequencers to play no note

  ShortestPossibleEnvelopeTime =      1;             // In samples
  EnvelopeSmoothing            =     99;             // envout := ( EnvelopeSmooting * OldValue + NewValue) / ( 1 + EnvelopeSmoothing); - was 49
  MinEnvTimeFast               =      0.0005;        // Minimum time for fast   rate envelopes in seconds
  MinEnvTimeMedium             =      0.05;          // Minimum time for medium rate envelopes in seconds
  MinEnvTimeSlow               =      5;             // Minimum time for slow   rate envelopes in seconds
  MaxEnvTimeFast               =     12;             // Maximum time for fast   rate envelopes in seconds
  MaxEnvTimeMedium             =   1200;             // Maximum time for medium rate envelopes in seconds
  MaxEnvTimeSlow               = 120000;             // Maximum time for slow   rate envelopes in seconds
  CompressorSmoothing          = 1.0 / 20.0;         // envelope smoothing for compressor - was 9 then was 19, now is a float

  {$IFDEF DEBUG}
    UNDEFINED = NAN;                                 // NAN for stability checks
    {$INLINE OFF}
  {$ELSE}
    UNDEFINED = 0.0;                                 // 0.0 for production code
    {$INLINE ON}
  {$ENDIF}

  scale_chromatic          =  0;
  scale_octatonic          =  1;
  scale_Istrian            =  2;
  scale_Altered            =  3;
  scale_Locrian            =  4;
  scale_Phrygian           =  5;
  scale_Neapolitan_minor   =  6;
  scale_Neapolitan_major   =  7;
  scale_Persian            =  8;
  scale_gypsy              =  9;
  scale_Phrygian_dominant  = 10;
  scale_Flamenco           = 11;
  scale_Tritone            = 12;
  scale_Enigmatic          = 13;
  scale_Iwato              = 14;
  scale_Insen              = 15;
  scale_blues_major        = 16;
  scale_Octatonic_2        = 17;
  scale_Half_diminished    = 18;
  scale_minor              = 19;
  scale_Harmonic_minor_1   = 20;
  scale_Dorian             = 21;
  scale_Melodic_minor      = 22;
  scale_harmonic_minor_2   = 23;
  scale_Hungarian_gyps     = 24;
  scale_Algerian           = 25;
  scale_Ukrainian_Dorian   = 26;
  scale_hungarian_gypsy    = 27;
  scale_hung_gyp_augmented = 28;
  scale_Hirajoshi          = 29;
  scale_Major_Locrian      = 30;
  scale_Major_bebop        = 31;
  scale_Adonai_malakh      = 32;
  scale_Harmonic_major     = 33;
  scale_Bebop_dominant     = 34;
  scale_Mixolydian         = 35;
  scale_major              = 36;
  scale_Acoustic           = 37;
  scale_Lydian             = 38;
  scale_Lydian_augmented   = 39;
  scale_whole_tone         = 40;
  scale_Prometheus         = 41;
  scale_pentatonic_Major   = 42;
  scale_Augmented          = 43;
  scale_blues_minor        = 44;
  scale_pentatonic_minor   = 45;
  scale_Yo                 = 46;
  scale_japanese           = 47;


var

  ReferenceA         : TSignal;  // The value of the reference A to be used - default 440 Hz
  NotesPerOctave     : TSignal;  // The number of notes wanted in an octave - there only is notation for 12 tho ... A .. G#
  MiddleNote         : TSignal;  // The middle note to be used - default 69 - this is a note number


  // -------------------------------------------------------------------------------------------------------------------
  // NOTE : the following four CAN NOT BE CHANGED at will, some relations need to be maintained
  // for this see **1 and please use the SetSystemRate function to change the System_Rate.
  // -------------------------------------------------------------------------------------------------------------------
  // These can not be moved to the implementation section as that will prevent inlining of
  // some functions. Also after changing these audio interfaces need to be re-initialized and lookuptables
  // for modules must be regenerated.

  System_Rate        : Integer;  // 44100.0;                          // samples / s                         ( audio   rate)
  Control_Decimation : Integer;  // 8;                                // Decimation used for slow tick stuff ( control rate)
  System_Rate_2      : TSignal;  // System_Rate * 2;                  // samples / s
  Control_Rate       : TSignal;  // System_Rate / Control_Decimation; // samples / s

  // END NOTE.
  // -------------------------------------------------------------------------------------------------------------------


type

  EKnobsConversions   = class( Exception);
  ENoDefaultConverter = class( EKnobsConversions);

  TSignalRate = (
    srControl,
    srAudio
  );

  TMidiSeptet = 0 .. 127;
  TMidiFilter = set of TMidiSeptet;

//  TMidiScale = record
//    FName   : string;
//    FFilter : TMidiFilter;
//  end;
//
//  TMidiScales = array of TMidiScale;

  TBaseNote  = 0 .. 11;
  TScale     = set of TMidiSeptet;

  TEnvRange  = ( erSlow, erMedium, erFast);


const

  TBaseNotes : TScale = [ 0 .. 11];

type

  TSignalFunction = function( const aValue: TSignal): TSignal;


  TSignals = class
  public
    FSignals : TSignalArray;
  private
    function    GetCount : Integer;
    function    GetSignal( anIndex: Integer): TSignal;
    procedure   SetSignal( anIndex: Integer; const aValue: TSignal);
  public
    constructor Create( aCount: Integer);
    destructor  Destroy;                                                            override;
    procedure   Clear;
    procedure   Erase;
    procedure   AddSignal( const aValue: TSignal);
    procedure   SetSignals( aStartIndex: Integer; const aValue: TSignals);
    function    Interpolate( const anIndex: TSignal): TSignal;                    // index [0, Count-1]
  public
    property    Count                    : Integer read GetCount;
    property    Signal[ anIndex: Integer]: TSignal read GetSignal write SetSignal; default;
  end;

  TSignalPair = record
    Left  : TSignal;
    Right : TSignal;
  end;

//  TComplexSignal = record
//    Real : TSignal;
//    Imag : TSignal;
//  end;

  TSignalPairs = array of TSignalPair;

  TSignalFifo = class
  private
    FSignals        : TSignalPairs;  // The data buffer
    FInp            : Integer;       // Current buffer insertion  point
    FOutp           : Integer;       // Current buffer extraction point
    FFreeCount      : Integer;       // Number of free data slots left
    FFillCount      : Integer;       // Number of filled data slots available
    FOverflows      : Integer;       // Number of times Append  was called while the buffer had no free space
    FUnderflows     : Integer;       // Number of tiems GetData was called while the buffer had no items in it
    FLastValue      : TSignalPair;   // To repeatedly return the last pair when no samples are available
    FCountOverflows : Boolean;       // Set to true to keep count of overflows
  private
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
    function    GetSize      : Integer;
  public
    constructor Create( aSize: Integer);
    destructor  Destroy;                                 override;
    procedure   Clear;
    procedure   Append( const aLeft, aRight: TSignal);   overload;
    procedure   Append( const aPair: TSignalPair);       overload;
    function    GetData: TSignalPair;
  public
    property    Overflows     : Integer read GetOverflows;
    property    Underflows    : Integer read GetUnderflows;
    property    FillCount     : Integer read FFillCount;
    property    FreeCount     : Integer read FFreeCount;
    property    Size          : Integer read GetSize;
    property    CountOverflows: Boolean read FCountOverflows write FCountOverflows  default True;
  end;

  TMedianCalculator = class
  private
    FSignals    : TSignals;      // The data buffer
    FSorted     : TSignalArray;  // The sorted buffer
    FInp        : Integer;       // Current buffer insertion  point
    FOutp       : Integer;       // Current buffer extraction point
    FIsSorted   : Boolean;
  private
    procedure   CheckSorted;
    function    GetCount   : Integer;
    function    GetLowest  : TSignal;
    function    GetMedian  : TSignal;
    function    GetHighest : TSignal;
  private
    procedure   QuickSort;                                              overload;
    procedure   QuickSort( const aValues: TSignalArray; L, R: Integer); overload;
  public
    constructor Create( aSize: Integer);
    destructor  Destroy;                                                override;
    procedure   Append( const aValue: TSignal);
    function    GetData: TSignal;
  public
    property    Count   : Integer read GetCount;
    property    Lowest  : TSignal read GetLowest;
    property    Median  : TSignal read GetMedian;
    property    Highest : TSignal read GetHighest;
  end;

  TRMSCalculator = class
  // http://www.embedded.com/design/configurable-systems/4006520/Improve-your-root-mean-calculations
  private
    FAvgTime       : TSignal;
    FMean          : TSignal;
    FRecipRootMean : TSignal;
    FAvgCoef       : TSignal;
  private
    procedure   Recalculate;
    procedure   SetAvgTime( aValue: TSignal);
  public
    constructor Create;
    procedure   Reset;
    procedure   Tick( anInValue: TSignal; var anOutValue: TSignal);
    procedure   SampleRateChanged;
  public
    property    AvgTime : TSignal read FAvgTime write SetAvgTime;
  end;

  TDbMinMax = class
  private
    FLowValue   : TSignal;
    FHighValue  : TSignal;
    FValeyDecay : TSignal;
    FPeakDecay  : TSignal;
    FValue      : TSignal;
    FPeak       : TSignal;
    FValey      : TSignal;
    FMin        : TSignal;
    FMax        : TSignal;
  public
    constructor Create( const aLowValue, aHighValue, aValeyDecay, aPeakDecay: TSignal);
    procedure   Tick( const aLinearValue: TSignal);
    procedure   Reset;
    procedure   Clear;
    function    ValueRanged( aMin, aMax: Integer): Integer;
    function    PeakRanged ( aMin, aMax: Integer): Integer;
    function    ValeyRanged( aMin, aMax: Integer): Integer;
  public
    property    LowValue   : TSignal read FLowValue  ;
    property    HighValue  : TSignal read FHighValue ;
    property    ValeyDecay : TSignal read FValeyDecay;
    property    PeakDecay  : TSignal read FPeakDecay ;
    property    Value      : TSignal read FValue     ;
    property    Peak       : TSignal read FPeak      ;
    property    Valey      : TSignal read FValey     ;
    property    Min        : TSignal read FMin       ;
    property    Max        : TSignal read FMax       ;
  end;

  TKnobsNullConverter = class
  private
    FSearched : string;
  protected
    FName     : string;
    FRate     : TSignalRate;
  private
    function    GetRate: TSignal;
  public
    constructor Create;                                                                  virtual;
    function    PosToHint     ( aPos, aSteps: Integer): string;                          virtual;
    function    PosToDisplay  ( aPos, aSteps: Integer): string;                          virtual;
    function    PosToValue    ( aPos, aSteps: Integer): TSignal;                         virtual;
  public
    property    Name     : string  read FName;
    property    Rate     : TSignal read GetRate;
    property    Searched : string  read FSearched write FSearched;
  end;

  TKnobsConverterClass = class of TKnobsNullConverter;

  TKnobsConverters = class( TStringList)
  private
    function    Converter( const anIndex: string): TKnobsNullConverter;
  public
    constructor Create;
    procedure   RegisterConverter( aClass: TKnobsConverterClass);
    procedure   RegisterConverters( const aClasses: array of TKnobsConverterClass);
    function    PosToHint   ( const aName: string; aPos, aSteps: Integer): string;
    function    PosToDisplay( const aName: string; aPos, aSteps: Integer): string;
    function    PosToValue  ( const aName: string; aPos, aSteps: Integer): TSignal;
    function    Rate        ( const aName: string): TSignal;
  end;

  function  IsTimeValue( const aValue: string): Boolean;
  function  ApplyUnitsToSignal( var aValue: TSignal; const aUnits: string; const aRate: TSignal): Boolean;
  function  ParseTextValue( const aString: string; var aSignal: TSignal; const aRate: TSignal): Boolean;

  function  ExpAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LinAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LogAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  ExpDecay ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LinDecay ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LogDecay ( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;

  function  FMod       ( const aValue, aDivisor: TSignal): TSignal; assembler; register;
  function  RangeMap   ( const aValue, anInLow, anInHigh, anOutLow, anOutHigh: TSignal): TSignal; inline;
  function  AlmostEqual( const aValue1, aValue2: TSignal; const aDelta: TSignal = DELTA): Boolean; inline;
  function  IsDenormal ( const aValue: TSignal): Boolean; inline;
  function  Normalize  ( const aValue : TSignal): TSignal; inline;
  procedure PNormalize ( var aValue : TSignal); inline;
  function  AlmostZero ( const aValue : TSignal; const aDelta: TSignal = DELTA): Boolean; inline;
  function  CubeInterp ( const fr, inm1, inp, inp1, inp2: TSignal): TSignal; inline;

  function  SignalTypeToColor( aValue: TSignalType): TColor;
  function  ColorToSignalType( aValue: TColor): TSignalType;

  function  LogicToSignal             ( const aValue : Boolean): TSignal; inline;
  function  SignalToLogic             ( const aValue : TSignal): Boolean; inline;
  function  BinaryToSignal            ( const aValue : TBinary; aBitCount: TBitCount): TSignal; inline;
  function  SignalToBinary            ( const aValue : TSignal; aBitCount: TBitCount): TBinary; inline;
  procedure TestBinary;
  function  SignalToMute              ( const aValue : TSignal): TSignal; inline;
  function  SignalToInt               ( const aValue : TSignal): Integer; inline;
  function  SignalTodB                ( const aValue : TSignal): TSignal; inline;
  function  dBToSignal                ( const aValue : TSignal): TSignal; inline;
  function  CentsToMultiplier         ( const aValue : TSignal): TSignal; inline;

  function  NoteNumberToUnits         ( const aValue : TSignal): TSignal; inline;
  function  UnitsToNoteNumber         ( const aValue : TSignal): TSignal; inline;
  function  UnitsToFrequency          ( const aValue : TSignal): TSignal; inline;
  function  UnitsToFrequencyLin       ( const aValue : TSignal): TSignal; inline;
  function  UnitsToFrequencyCubed     ( const aValue : TSignal): TSignal; inline;
  function  FrequencyToUnits          ( const aValue : TSignal): TSignal; inline;
  function  FrequencyToPhaseDelta     ( const aValue : TSignal): TSignal; inline;
  function  UnitsToPhaseDelta         ( const aValue : TSignal): TSignal; inline;
  function  UnitsToPhaseDeltaLin      ( const aValue : TSignal): TSignal; inline;
  function  SampleCountToTime         ( const aValue : TSignal): TSignal; inline;
  function  TimeToSampleCount         ( const aValue : TSignal): TSignal; inline;

  function  UnitsToLfoFastFrequency   ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoMediumFrequency ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoSlowFrequency   ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoBPMFrequency    ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoLinFrequency    ( const aValue : TSignal): TSignal; inline;
  function  FrequencyToLfoPhaseDelta  ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoFastPhaseDelta  ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoMediumPhaseDelta( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoSlowPhaseDelta  ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoBPMPhaseDelta   ( const aValue : TSignal): TSignal; inline;
  function  UnitsToLfoLinPhaseDelta   ( const aValue : TSignal): TSignal; inline;
  function  SlowTimeToSampleCount     ( const aValue : TSignal): TSignal; inline;

  function  UnitsToSvfFrequency       ( const aValue : TSignal): TSignal;

  function  TimeToSeconds    ( const aValue    : string; const aDefault: TSignal): TSignal;
  function  NoteNameToUnits  ( const aNoteName : string; const aDefault: TSignal): TSignal;
  function  UnitsToNoteName  ( const aValue: TSignal; Compact: Boolean = True): string;
  function  UnitsToNoteName12( const aValue: TSignal): string;
  function  IsNoteName       ( const aNoteName: string): Boolean;

  function  SignalToStr ( const aValue: TSignal): string;

  function  TimeToStr   ( const aValue: TSignal): string;
  function  FreqToStr   ( const aValue: TSignal): string;
  function  FreqToBPMStr( const aValue: TSignal): string;

  function  ScaleIndexToData( aScale: Integer): TScale;
  function  ScaleIndexToName( aScale: Integer): string;

  function  MakeSignalPair   ( aLeft, aRight: TSignal): TSignalPair; inline;
  function  MakeComplexSignal( aReal, anImag: TSignal): TComplex;    inline;

  procedure RegisterConverters( const aClasses: array of TKnobsConverterClass);
  function  Converters: TKnobsConverters;


  function  LookupFrequency          ( const aValue: TSignal): TSignal;
  function  LookupFrequencyLin       ( const aValue: TSignal): TSignal;
  function  LookupFrequencyCubed     ( const aValue: TSignal): TSignal;
  function  LookupPhaseDelta         ( const aValue: TSignal): TSignal;
  function  LookupPhaseDeltaLin      ( const aValue: TSignal): TSignal;
  function  LookupSvfFrequency       ( const aValue: TSignal): TSignal;
  function  LookupLfoSlowFrequency   ( const aValue: TSignal): TSignal;
  function  LookupLfoSlowPhaseDelta  ( const aValue: TSignal): TSignal;
  function  LookupLfoMediumFrequency ( const aValue: TSignal): TSignal;
  function  LookupLfoMediumPhaseDelta( const aValue: TSignal): TSignal;
  function  LookupLfoFastFrequency   ( const aValue: TSignal): TSignal;
  function  LookupLfoFastPhaseDelta  ( const aValue: TSignal): TSignal;
  function  LookupLfoBPMFrequency    ( const aValue: TSignal): TSignal;
  function  LookupLfoBPMPhaseDelta   ( const aValue: TSignal): TSignal;
  function  LookupLfoLinFrequency    ( const aValue: TSignal): TSignal;
  function  LookupLfoLinPhaseDelta   ( const aValue: TSignal): TSignal;
  function  LookupSine               ( const aValue: TSignal): TSignal;
  function  LookupCosine             ( const aValue: TSignal): TSignal;
  function  LookupScaledEnvTime      ( const aValue: TSignal): TSignal;

  function  LookupEnvTime            ( aRange: TEnvRange; aValue: TSignal): TSignal;

  function  MidiFilterToStr( const aValue: TMidiFilter): string;

  procedure SetSystemRate( aValue: Integer);




implementation



var

  LocConverters : TKnobsConverters;



{ ========
  TSignals = class
  public
    FSignals : TSignalArray;
  public
    property    Count                    : Integer read GetCount;
    property    Signal[ anIndex: Integer]: TSignal read GetSignal write SetSignal; default;
  private
}

    function    TSignals.GetCount: Integer;
    begin
      Result := Length( FSignals);
    end;

    function    TSignals.GetSignal( anIndex: Integer): TSignal;
    begin
      Result := FSignals[ anIndex];
    end;

    procedure   TSignals.SetSignal( anIndex: Integer; const aValue: TSignal);
    begin
      FSignals[ anIndex] := aValue;
    end;

//  public

    constructor TSignals.Create( aCount: Integer); // overload;
    begin
      inherited Create;
      SetLength( FSignals, aCount);
    end;

    destructor  TSignals.Destroy; // override;
    begin
      Clear;
      inherited;
    end;

    procedure   TSignals.Clear;
    begin
      SetLength( FSignals, 0);
    end;

    procedure   TSignals.Erase;
    var
      i : Integer;
    begin
      for i := 0 to Length( FSignals) - 1
      do FSignals[ i] := 0.0;
    end;

    procedure   TSignals.AddSignal( const aValue: TSignal);
    begin
      SetLength( FSignals, Count + 1);
      FSignals[ Count - 1] := aValue;
    end;

    procedure   TSignals.SetSignals( aStartIndex: Integer; const aValue: TSignals);
    var
      i : Integer;
    begin
      if Assigned( aValue)
      then begin
        SetLength( FSignals, aStartIndex + aValue.Count);
        for i := aStartIndex to Count - 1
        do Signal[ i] := aValue.Signal[ i - aStartIndex];
      end;
    end;

    function    TSignals.Interpolate( const anIndex: TSignal): TSignal;
    var
      p : Integer;
      q : Integer;
      d  : TSignal;
    begin
      p := Trunc( anIndex);
      d := anIndex - p;
      q := ( p + 1);
      if q > Count - 1
      then q := 0;
      Result := FSignals[ p] + d * ( FSignals[ q] - FSignals[ p]);
    end;



{ ========
  TSignalPair = record
    Left  : TSignal;
    Right : TSignal;
  end;

  TSignalPairs = array of TSignalPair;

  TSignalFifo = class
  private
    FSignals        : TSignalPairs;  // The data buffer
    FInp            : Integer;       // Current buffer insertion  point
    FOutp           : Integer;       // Current buffer extraction point
    FFreeCount      : Integer;       // Number of free data slots left
    FFillCount      : Integer;       // Number of filled data slots available
    FOverflows      : Integer;       // Number of times Append  was called while the buffer had no free space
    FUnderflows     : Integer;       // Number of tiems GetData was called while the buffer had no items in it
    FLastValue      : TSignalPair;   // To repeatedly return the last pair when no samples are available
    FCountOverflows : Boolean;       // Set to true to keep count of overflows
  public
    property    Overflows     : Integer read GetOverflows;
    property    Underflows    : Integer read GetUnderflows;
    property    FillCount     : Integer read FFillCount;
    property    FreeCount     : Integer read FFreeCount;
    property    Size          : Integer read GetSize;
    property    CountOverflows: Boolean read FCountOverflows write FCountOverflows  default True;
  private
}

    function    TSignalFifo.GetOverflows: Integer;
    begin
      Result     := FOverflows;
      FOverflows := FOverflows - Result;
    end;

    function    TSignalFifo.GetUnderflows: Integer;
    begin
      Result      := FUnderflows;
      FUnderflows := FUnderflows - Result;
    end;

    function    TSignalFifo.GetSize: Integer;
    begin
      Result := Length( FSignals);
    end;

//  public

    constructor TSignalFifo.Create( aSize: Integer);
    begin
      inherited Create;
      CountOverflows := True;
      SetLength( FSignals, aSize);
      Clear;
    end;

    destructor  TSignalFifo.Destroy; // override;
    begin
      Clear;
      FFreeCount := 0;
      FFillCount := 0;
      inherited;
    end;

    procedure   TSignalFifo.Clear;
    begin
      FFillCount       := 0;
      FFreeCount       := 0;
      FInp             := 0;
      FOutp            := 0;
      FUnderflows      := 0;
      FOverflows       := 0;
      FLastValue.Left  := 0;
      FLastValue.Right := 0;
      FFreeCount       := Size;
    end;

    procedure   TSignalFifo.Append( const aLeft, aRight: TSignal); // overload;
    var
      aPair: TSignalPair;
    begin
      aPair.Left  := aLeft;
      aPair.Right := aRight;
      Append( aPair);
    end;

    procedure   TSignalFifo.Append( const aPair: TSignalPair); // overload;
    // When the queue is full discard sample pair, and incremeent the oveflow counteer
    begin
      if FFreeCount > 0
      then begin
        Dec( FFreeCount);
        FSignals[ FInp] := aPair;
        FInp := ( FInp + 1) mod Size;
        Inc( FFillCount);
      end
      else begin
        if CountOverflows
        then {$Q-R-} Inc( FOverflows) {$Q+R+};
      end;
    end;

    function    TSignalFifo.GetData: TSignalPair;
    // When no sample pairs are available repeatedly return tha last pair, and increment the underflow counter.
    begin
      if FFillCount > 0
      then begin
        Dec( FFillCount);
        FLastValue := FSignals[ FOutp];
        FOutp      := ( FOutp + 1) mod Size;
        Inc( FFreeCount);
      end
      else begin
        if CountOverflows
        then {$Q-R-} Inc( FUnderflows) {$Q+R+};
      end;
      Result := FLastValue;
    end;


{ ========
  TMedianCalculator = class
  private
    FSignals    : TSignals;      // The data buffer
    FSorted     : TSignalArray;  // The sorted buffer
    FInp        : Integer;       // Current buffer insertion  point
    FOutp       : Integer;       // Current buffer extraction point
    FIsSorted   : Boolean;
  public
    property    Count   : Integer read GetCount;
    property    Lowest  : TSignal read GetLowest;
    property    Median  : TSignal read GetMedian;
    property    Highest : TSignal read GetHighest;
  private
}

    procedure   TMedianCalculator.CheckSorted;
    var
      i : Integer;
    begin
      if ( Count > 0) and not FIsSorted
      then begin
        for i := 0 to Count - 1
        do FSorted[ i] := FSignals[ i];
        QuickSort;
        FIsSorted := True;
      end;
    end;

    function    TMedianCalculator.GetCount: Integer;
    begin
      if Assigned( FSignals)
      then Result := FSignals.Count
      else Result := 0;
    end;

    function    TMedianCalculator.GetLowest: TSignal;
    begin
      if Count > 0
      then begin
        CheckSorted;
        Result := FSorted[ 0];
      end
      else Result := 0;
    end;

    function    TMedianCalculator.GetMedian: TSignal;
    begin
      if Count > 0
      then begin
        CheckSorted;
        if Odd( Count)
        then Result := FSorted[ Count div 2]
        else Result := ( FSorted[ Count div 2 - 1] + FSorted[ Count div 2]) * 0.5;
      end
      else Result := 0;
    end;

    function    TMedianCalculator.GetHighest: TSignal;
    begin
      if Count > 0
      then begin
        CheckSorted;
        Result := FSorted[ Count - 1];
      end
      else Result := 0;
    end;

//  private

    procedure   TMedianCalculator.QuickSort; // overlaod;
    begin
      QuickSort( FSorted, 0, Count - 1);
    end;

    procedure   TMedianCalculator.QuickSort( const aValues: TSignalArray; L, R: Integer); // overlaod;
    var
      I, J, P: Integer;
      aTmp : TSignal;
    begin
      repeat
        I := L;
        J := R;
        P := ( L + R) shr 1;
        repeat
          while aValues[ I] < aValues[ P] do Inc( I);
          while aValues[ J] > aValues[ P] do Dec( J);
          if I <= J
          then begin
            if I <> J
            then begin
              aTmp        := aValues[ I];
              aValues[ I] := aValues[ J];
              aValues[ J] := aTmp;
            end;
            if P = I
            then P := J
            else if P = J
            then P := I;
            Inc( I);
            Dec( J);
          end;
        until I > J;
        if L < J then QuickSort( aValues, L, J);
        L := I;
      until I >= R;
    end;

//  public

    constructor TMedianCalculator.Create( aSize: Integer);
    begin
      inherited Create;
      FSignals := TSignals.Create( aSize);
      SetLength( FSorted, aSize);
    end;

    destructor  TMedianCalculator.Destroy; // override;
    begin
      FreeAndNil( FSignals);
      inherited;
    end;

    procedure   TMedianCalculator.Append( const aValue: TSignal);
    begin
      FSignals[ FInp] := aValue;
      FInp            := ( FInp + 1) mod FSignals.Count;
      FIsSorted       := False;
    end;

    function    TMedianCalculator.GetData: TSignal;
    begin
      Result := FSignals[ FOutp];
      FOutp  := ( FOutp + 1) mod FSignals.Count;
    end;

{ ========
  TRMSCalculator = class
  private
    FAvgTime       : TSignal;
    FMean          : TSignal;
    FRecipRootMean : TSignal;
    FAvgCoef       : TSignal;
  public
    property    AvgTime : TSignal read FAvgTime write SetAvgTime;
  orivate
}

    procedure   TRMSCalculator.Recalculate;
    begin;
      FAvgCoef := 1.0 - Exp( - 1.0 / TimeToSampleCount( FAvgTime));
    end;

    procedure   TRMSCalculator.SetAvgTime( aValue: TSignal);
    begin
      FAvgTime := aValue;
      Recalculate;
    end;

//  public

    constructor TRMSCalculator.Create;
    begin
      AvgTime    := 0.1;              // 100 ms
      Reset;
    end;

    procedure   TRMSCalculator.Reset;
    begin
      FMean          := 0.0;
      FRecipRootMean := 1e-10;        // 0 < FRecipRootMean < 1
    end;

    procedure   TRMSCalculator.Tick( anInValue: TSignal; var anOutValue: TSignal);
    begin
      FMean          := FMean + FAvgCoef * ( anInValue - FMean);
      FRecipRootMean := FRecipRootMean * 0.5 * ( 3.0 - ( FRecipRootMean * FRecipRootMean * FMean));
      anOutValue     := FRecipRootMean * FMean;
    end;

    procedure   TRMSCalculator.SampleRateChanged;
    // Notification for sample rate change, the actual values are in global variables
    // System_Rate for the audio rate and Control_Rate for the control rate
    // Modules caching sample rate dependent data should recalculate sych data
    // in an overriden version of this procedure.
    begin
      Recalculate;
    end;


{ ========
  TDbMinMaxCalculator = class
  private
    FLowValue   : TSignal;
    FHighValue  : TSignal;
    FValeyDecay : TSignal;
    FPeakDecay  : TSignal;
    FValue      : TSignal;
    FPeak       : TSignal;
    FValey      : TSignal;
    FMin        : TSignal;
    FMax        : TSignal;
  public
    property    LowValue   : TSignal read FLowValue  ;
    property    HighValue  : TSignal read FHighValue ;
    property    ValeyDecay : TSignal read FValeyDecay;
    property    PeakDecay  : TSignal read FPeakDecay ;
    property    Value      : TSignal read FValue     ;
    property    Peak       : TSignal read FPeak      ;
    property    Valey      : TSignal read FValey     ;
    property    Min        : TSignal read FMin       ;
    property    Max        : TSignal read FMax       ;
  public
}

    constructor TDbMinMax.Create( const aLowValue, aHighValue, aValeyDecay, aPeakDecay: TSignal);
    begin
      inherited Create;
      FLowValue   := aLowValue;
      FHighValue  := aHighValue;
      FValeyDecay := aValeyDecay;
      FPeakDecay  := aPeakDecay;
      Clear;
    end;

    procedure   TDbMinMax.Tick( const aLinearValue: TSignal);
    begin
      FValue := SignalToDb( aLinearValue);

      if FValue > FPeak
      then FPeak := FValue
      else FPeak := FPeakDecay * ( FPeak - FValue) + FValue;

      if FValue < FValey
      then FValey := FValue
      else FValey := FValeyDecay * ( FValey - FValue) + FValue;

      if FValue > FMax
      then FMax := FValue;

      if FValue < FMin
      then FMin := FValue;
    end;

    procedure   TDbMinMax.Reset;
    begin
      FPeak  := FValue;
      FValey := FValue;
      FMin   := FValue;
      FMax   := FValue;
    end;

    procedure   TDbMinMax.Clear;
    begin
      FMin   := FHighValue;
      FMax   := FLowValue;
      FValue := FLowValue;
      FPeak  := FValue;
      FValey := FValue;
    end;

    function    TDbMinMax.ValueRanged( aMin, aMax: Integer): Integer;
    begin
      Result := Round( RangeMap( Clip( FValue, FLowValue, FHighValue), FLowValue, FHighValue, aMin, aMax));
    end;

    function    TDbMinMax.PeakRanged( aMin, aMax: Integer): Integer;
    begin
      Result := Ceil( RangeMap( Clip( FPeak, FLowValue, FHighValue), FLowValue, FHighValue, aMin, aMax));
    end;

    function    TDbMinMax.ValeyRanged( aMin, aMax: Integer): Integer;
    begin
      Result := Floor( RangeMap( Clip( FValey, FLowValue, FHighValue), FLowValue, FHighValue, aMin, aMax));
    end;



  function  IsTimeValue( const aValue: string): Boolean;
  begin
    Result := Pos( ':', aValue) > 0;
  end;

  function  ApplyUnitsToSignal( var aValue: TSignal; const aUnits: string; const aRate: TSignal): Boolean;
  begin
    Result := True;
    if not ISNan( aValue)
    then aValue := aValue * System_Rate / aRate;
    if      SameText( aUnits, '')
    then // nothing
    else if SameText( aUnits, 'Hz')
    then aValue := FrequencyToUnits( aValue)
    else if SameText( aUnits, 'kHz')
    then aValue := FrequencyToUnits( 1000 * aValue)
    else if SameText( aUnits, 'mHz')
    then aValue := FrequencyToUnits( aValue / 1000)
    else if SameText( aUnits, 'bpm')
    then aValue := FrequencyToUnits( aValue / 60.0)
    else if SameText( aUnits, 'degrees') or SameText( aUnits, 'deg')
    then aValue := aValue / 360
    else if SameText( aUnits, 'radians') or SameText( aUnits, 'rad')
    then aValue := aValue / TWO_PI
    else if SameText( aUnits, 'percent') or SameText( aUnits, '%')
    then aValue := aValue / 100
    else if SameText( aUnits, 'ms')
    then aValue := aValue / 1000
    else if SameText( aUnits, 's')
    then // nothing
    else if SameText( aUnits, 'nn')
    then aValue := NoteNumberToUnits( aValue)
    else if SameText( aUnits, 'qsteps') or SameText( aUnits, 'q')
    then aValue := aValue / 1000
    else if SameText( aUnits, 'ct')
    then aValue := aValue / 256
    else Result := False;
  end;

  function  ParseTextValue( const aString: string; var aSignal: TSignal; const aRate: TSignal): Boolean;
  var
    aParts  : TStringList;
    S       : string;
    aUnits  : string;
  begin
    Result := True;
    aParts := Explode( aString, ' ');
    try
      aUnits := '';
      if aParts.Count > 0
      then begin
        S := Trim( aParts[ 0]);
        if IsNoteName( S)
        then aSignal := NoteNameToUnits  ( S, UNDEFINED)
        else if IsTimeValue( S)
        then aSignal := 1 / TimeToSeconds( S, UNDEFINED)
        else aSignal := StrToFloatDef    ( S, UNDEFINED);
      end
      else Result := False;
      if Result and ( aParts.Count > 1)
      then begin
        aUnits := Trim( aParts[ 1]);
        Result := ApplyUnitsToSignal( aSignal, aUnits, aRate);
      end;
    finally
      aParts.Free;
    end;
  end;

  function  ExpAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 ^ x - 1  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := LookupScaledEnvTime( aTim);
    Result   := Normalize(( 1 - Unscaled) * aStartValue + Unscaled * anEndValue);
  end;

  function  LinAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = x  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := aTim;
    Result   := Normalize(( 1 - Unscaled) * aStartValue + Unscaled * anEndValue);
  end;

  function  LogAttack( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 - 2 ^ ( 1 - x)  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := 1 - LookupScaledEnvTime( 1 - aTim);
    Result   := Normalize(( 1 - Unscaled) * aStartValue + Unscaled * anEndValue);
  end;

  function  ExpDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 ^ ( 1 - x) - 1  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := LookupScaledEnvTime( 1 - aTim);
    Result   := Normalize(( 1 - Unscaled) * anEndValue + Unscaled * aStartValue);
  end;

  function  LinDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 1 - x  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := 1 - aTim;
    Result   := Normalize(( 1 - Unscaled) * anEndValue + Unscaled * aStartValue);
  end;

  function  LogDecay( const aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 - 2 ^ x  [0,1] -> [0,1]
  var
    Unscaled : TSignal;
    aDur     : TSignal;
    aTim     : TSignal;
  begin
    aDur     := Max( 1, aDuration);
    aTim     := Clip( aTime, 0, aDur) / aDur;
    Unscaled := 1 - LookupScaledEnvTime( aTim);
    Result   := Normalize(( 1 - Unscaled) * anEndValue + Unscaled * aStartValue);
  end;

  function  FMod( const aValue, aDivisor: TSignal): TSignal; assembler; register;
  asm
    fld qword ptr[ aDivisor]
    fld qword ptr[ aValue  ]
@r:
    fprem
    fstsw ax
    sahf
    jp    @r
    fstp  st( 1)
  end;

  function  IsDenormal( const aValue: TSignal): Boolean; inline;
  // http://musicdsp.org/showArchiveComment.php?ArchiveID=88
  begin
    Result := ( PCardinal( @ aValue)^ and $7fe00000) <> 0;
  end;

  function  Normalize( const aValue: TSignal): TSignal; inline;
  begin
//    Result := aValue + 1e-20 - 1e-20;
    if Abs( aValue) < 1e-20
    then Result := 0
    else Result := aValue;
  end;

  procedure PNormalize( var aValue : TSignal); inline;
  begin
    if Abs( aValue) < 1e-20
    then aValue := 0;
  end;

  function  RangeMap( const aValue, anInLow, anInHigh, anOutLow, anOutHigh: TSignal): TSignal; inline;
  begin
    Result := anOutLow + (( aValue - anInLow) / ( anInHigh - anInLow)) * ( anOutHigh - anOutLow);
  end;

  function  AlmostEqual( const aValue1, aValue2: TSignal; const aDelta: TSignal = DELTA): Boolean;  inline;
  begin
    Result := Abs( aValue1 - aValue2) < aDelta;
  end;

  function  AlmostZero( const aValue: TSignal; const aDelta: TSignal = DELTA): Boolean; inline;
  begin
    Result := Abs( aValue) < aDelta;
  end;

  function  CubeInterp( const fr, inm1, inp, inp1, inp2: TSignal): TSignal; inline;
  begin
    Result :=
      inp + 0.5 * fr * ( inp1 - inm1 + fr * ( 4.0 * inp1 + 2.0 * inm1 - 5.0 * inp - inp2 + fr * ( 3.0 * ( inp - inp1) - inm1 + inp2)));
  end;

  function  SignalTypeToColor( aValue: TSignalType): TColor;
  begin
    case aValue of
      stControl : Result := clBlue;
      stAudio   : Result := clRed;
      stLogic   : Result := clYellow;
      else        Result := clWhite;
    end;
  end;

  function  ColorToSignalType( aValue: TColor): TSignalType;
  begin
    case aValue of
      clBlue   : Result := stControl;
      clRed    : Result := stAudio;
      clYellow : Result := stLogic;
      else       Result := stUnknown;
    end;
  end;

  function  LogicToSignal( const aValue: Boolean): TSignal; inline;
  begin
    if aValue
    then Result := 1
    else Result := -1;
  end;

  function  SignalToLogic( const aValue: TSignal): Boolean; inline;
  begin
    Result := aValue > 0;
  end;

  function  BinaryToSignal( const aValue : TBinary; aBitCount: TBitCount): TSignal; inline;
  // done : make a LUT. - used shifts instead
  begin
    Result := ( 1.0 * aValue) / ( 1 shl ( aBitCount - 1));
//  Result := ( 1.0 * aValue) / Power( 2, aBitCount - 1);
  end;

  function  SignalToBinary( const aValue : TSignal; aBitCount: TBitCount): TBinary; inline;
  // done : make a LUT. - used shifts instead
  begin
    Result := Round( Clip( aValue, -1, 1) * ( 1 shl ( aBitCount - 1)));
//  Result := Round( Clip( aValue, -1, 1) * Power( 2, aBitCount - 1));
  end;

  procedure TestBinary;
  // done : make a LUT. - used shifts instead - used pre-computed shifts instead
  {$IFDEF DEBUG}
  const
    BitCount : Integer = 32;
    BitSize  : Integer = 1 shl ( 32 - 1);
  var
    i : TBinary;
    V : TSignal;
    L : Integer;
    R : Integer;
  {$ENDIF}
  begin
    {$IFDEF DEBUG}
    L := Round( BitSize + 0.4999) - 1;
//  L := Round(( 1 shl ( BitCount - 1)) + 0.4999) - 1;
//  L := Round( Power( 2, BitCount - 1) + 0.4999) - 1;
    for i := -L to L
    do begin
      V := BinaryToSignal( i, BitCount);
      R := SignalToBinary( V, BitCount);
      if i <> R
      then Nop;
    end;
    {$ENDIF}
  end;

  function  SignalToMute( const aValue  : TSignal): TSignal; inline;
  begin
    Result := Normalize( 1 - aValue);
  end;

  function  SignalToInt( const aValue : TSignal): Integer; inline;
  begin
    Result := Round( aValue + 0.4999);
  end;

  function  SignalTodB( const aValue : TSignal): TSignal; inline;
  // todo : make a LUT.
  begin
    if aValue <= 1e-7
    then Result := -140
    else Result := 20 * Log10( aValue);
  end;

  function  dBToSignal( const aValue : TSignal): TSignal; inline;
  // todo : make a LUT.
  begin
    Result := Power( 10, aValue / 20);
  end;

  function  CentsToMultiplier( const aValue : TSignal): TSignal; inline;
  // todo : make a LUT.
  begin
    Result := Power( Power( 2, aValue / 100.0), 1.0 / 12.0);
  end;

  function  UnitsToNoteNumber( const aValue: TSignal): TSignal; inline;
  begin
    Result := 128 * aValue;
  end;

  function  NoteNumberToUnits( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue / 128;
  end;

  function  UnitsToFrequency( const aValue: TSignal): TSignal; inline;
  // such that note Nr 69 maps to A 440 Hz
  // see: http://www.phys.unsw.edu.au/jw/notes.html
  // freq = ( 2 ^ ( notenr - 69) / 12) * 440
  // Result := Power( 2, ( aValue - 69) / 12) * ReferenceA;
  begin
    Result := Power( 2, ( UnitsToNoteNumber( aValue) - MiddleNote) / NotesPerOctave) * ReferenceA;
  end;

  function  UnitsToFrequencyLin( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * 10000;
  end;

  function  UnitsToFrequencyCubed( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * aValue * aValue * 10000;
  end;

  function  FrequencyToUnits( const aValue: TSignal): TSignal; inline;
  // see: http://www.phys.unsw.edu.au/jw/notes.html
  // n  = 12 * log2( f / 440) + 69
  //  Result := Round( 12 * Log2( aValue / ReferenceA)) + 69;
  begin
    Result := NoteNumberToUnits( NotesPerOctave * Log2( aValue / ReferenceA) + MiddleNote);
  end;

  function  FrequencyToPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue / System_Rate;
  end;

  function  UnitsToPhaseDelta( const aValue : TSignal): TSignal; inline;
  begin
    Result := FrequencyToPhaseDelta( UnitsToFrequency( aValue));
  end;

  function  UnitsToPhaseDeltaLin( const aValue : TSignal): TSignal; inline;
  begin
    Result := FrequencyToPhaseDelta( UnitsToFrequencyLin( aValue));
  end;

  function  SampleCountToTime( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue / System_Rate;
  end;

  function  TimeToSampleCount( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * System_Rate;
  end;

  // ------------------------------------

  function  UnitsToLfoFastFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Power( 2, ( UnitsToNoteNumber( aValue) - 43) / 12) * 3.05);
  end;

  function  UnitsToLfoMediumFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Power( 2, ( UnitsToNoteNumber( aValue) - 103) / 12) * 3.05);
  end;

  function  UnitsToLfoSlowFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Power( 2, ( UnitsToNoteNumber( aValue) - 163) / 12) * 3.05);
  end;

  function  UnitsToLfoBPMFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( UnitsToNoteNumber( aValue) / 30);
  end;

  function  UnitsToLfoLinFrequency( const aValue: TSignal): TSignal; // inline;
  begin
    Result := aValue * 15;
  end;

  function  FrequencyToLfoPhaseDelta( const aValue: TSignal): TSignal; // inline;
  begin
    Result := Normalize( Control_Decimation * aValue / System_Rate);
  end;

  function  UnitsToLfoFastPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoFastFrequency( aValue));
  end;

  function  UnitsToLfoMediumPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoMediumFrequency( aValue));
  end;

  function  UnitsToLfoSlowPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoSlowFrequency( aValue));
  end;

  function  UnitsToLfoBPMPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoBPMFrequency( aValue));
  end;

  function  UnitsToLfoLinPhaseDelta( const aValue: TSignal): TSignal; inline;
  begin
    Result := FrequencyToLfoPhaseDelta( UnitsToLfoLinFrequency( aValue))
  end;

  function  SlowTimeToSampleCount( const aValue: TSignal): TSignal; inline;
  begin
    Result := aValue * Control_Rate;
  end;

  function  UnitsToSvfFrequency( const aValue: TSignal): TSignal;
  // NOTE: this is based on an SVF sampled at twice the system rate
  begin
    Result := Normalize( 2 * Sin( PI * Min( 0.25, ( UnitsToFrequency( aValue) / System_Rate_2))));
  end;

  function  UnitsToSine( const aValue: TSignal): TSignal;
  begin
    Result := Normalize( Sin( TWO_PI * aValue));
  end;

  function  UnitsToCosine( const aValue: TSignal): TSignal;
  begin
    Result := Normalize( Cos( TWO_PI * aValue));
  end;

  function  UnitsToScaledEnvTime( const aValue: TSignal): TSignal;
  begin
    Result := Normalize(( Power( 2, 10 * aValue) - 1 ) / 1023);
  end;


  function  TimeToSeconds( const aValue: string; const aDefault: TSignal): TSignal;
  var
    aParts   : TStringList;
    anHours  : Integer;
    aMinutes : Integer;
    aSeconds : Integer;
  begin
    Result    := aDefault;
    anHours   := -1;
    aMinutes  := -1;
    aSeconds  := -1;
    aParts := Explode( aValue, ':');
    try
      if aParts.Count = 2
      then begin
        aSeconds := StrToIntDef( aParts[ 1], -1);
        aMinutes := StrToIntDef( aParts[ 0], -1);
        anHours  := 0;
      end
      else if aParts.Count = 3
      then begin
        aSeconds := StrToIntDef( aParts[ 2], -1);
        aMinutes := StrToIntDef( aParts[ 1], -1);
        anHours  := StrToIntDef( aParts[ 0], -1);
      end;
      if ( aSeconds >= 0) and ( aMinutes >= 0) and ( anHours >= 0)
      then Result := aSeconds + 60 * ( aMinutes + 60 * anHours);
    finally
      aParts.Free;
    end;
  end;

  function  NoteNameToUnits( const aNoteName: string; const aDefault: TSignal): TSignal;
  // C[#][-]0
  var
    Base  : Integer;
    Sharp : Integer;
    Sign  : Integer;
    N     : Char;
    Shift : Integer;
    p     : Integer;
    T     : string;
    NN    : Integer;
  begin
    Shift :=  0;
    Sharp :=  0;
    Sign  :=  1;
    Base  := -1;
    if Length( aNoteName) > 0
    then begin
      N := UpCase( aNoteName[ 1]);
      case N of
        'C' : Base := 12     ;
        'D' : Base := 12 +  2;
        'E' : Base := 12 +  4;
        'F' : Base := 12 +  5;
        'G' : Base := 12 +  7;
        'A' : Base := 12 +  9;
        'B' : Base := 12 + 11;
        else  Base := -1;
      end;
      if ( Length( aNoteName) > 1) and ( Base > 0)
      then begin
        if aNoteName[ 2] = '#'
        then begin
          Sharp := 1;
          p     := 2;
        end
        else p := 1;
      end
      else p := 0;
      if ( Length( aNoteName) > p) and ( base > 0)
      then begin
        if aNoteName[ p + 1] = '-'
        then begin
          Sign := -1;
          p    := p + 1;
        end;
      end;
      T := Copy( aNoteName, p + 1, Length( aNoteName));
      if T[ Length( T)] = '#'
      then begin
        Sharp := 1;
        T     := Copy( T, 1, Length( T) - 1);
      end;
      Shift := StrToIntDef( T, 0);
    end;
    if Base > 0
    then begin
      NN     := Base + Sharp + Sign * Shift * 12;
      Result := NoteNumberToUnits( NN);
    end
    else Result := aDefault;
  end;

  function  UnitsToNoteName( const aValue: TSignal; Compact: Boolean = True): string;
  const
    Notes  : array[ 0 .. 11] Of Char = 'CCDDEFFGGAAB';
    Sharps : array[ 0 .. 11] Of Char = ' # #  # # # ';
  var
    i          : Integer;
    S          : string;
    aNoteValue : Integer;
  begin
    if IsNan( aValue)
    then aNoteValue := Round( MiddleNote)
    else aNoteValue := Round( UnitsToNoteNumber( aValue));
    i          := IntMod( aNoteValue, 12);
    S          := Sharps[ i];
    if Compact and ( S = ' ')
    then S := '';
    Result := Format( '%s%s%d', [ Notes[ i], S, ( IntDiv( aNoteValue, 12)) - 1], AppLocale);
  end;

  function  UnitsToNoteName12( const aValue: TSignal): string;
  const
    Names : array[ TBaseNote] of string = (
      'C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'
    );
  var
    aNote : Integer;
  begin
    aNote := SignalToInt( 12 * aValue);
    if aNote in TBaseNotes
    then Result := Names[ aNote]
    else Result := 'inv';
  end;

  function  IsNoteName( const aNoteName: string): Boolean;
  // C[#][-]0[#]
  var
    p : Integer;
    i : Integer;
  begin
    Result := False;
    if Length( aNoteName) > 0
    then begin
      Result := CharInSet( aNoteName[ 1], ['a' .. 'g', 'A' .. 'G']);
      if Result and ( Length( aNoteName) > 1)
      then begin
        if aNoteName[ 2] = '#'
        then p := 2
        else p := 1;
      end
      else p := 0;

      if Result and ( Length( aNoteName) > p) and ( aNoteName[ p + 1] = '-')
      then p := p + 1;

      if Result
      then begin
        for i := p + 1 to Length( aNoteName)
        do begin
          if not CharInset( aNoteName[ i], [ '0' .. '9'])
          then begin
            if ( aNoteName[ i] = '#') and ( i = Length( aNoteName))
            then Break
            else begin
              Result := False;
              Break;
            end;
          end;
        end;
      end;
    end;
  end;

  function  SignalToStr( const aValue: TSignal): string;
  begin
    Result := FloatToStrF( aValue, ffGeneral, 15, 0, AppLocale);
  end;

  function  TimeToStr( const aValue: TSignal): string;
  begin
    if aValue = 0
    then Result := '0 s'
    else if aValue < 0.0001
    then Result := Format( '%.1f s', [ aValue * 1e6], AppLocale)
    else if aValue < 0.001
    then Result := Format( '%.0f s', [ aValue * 1e6], AppLocale)
    else if aValue < 0.01
    then Result := Format( '%.2f ms', [ aValue * 1e3], AppLocale)
    else if aValue < 0.1
    then Result := Format( '%.1f ms', [ aValue * 1e3], AppLocale)
    else if aValue < 1
    then Result := Format( '%.0f ms', [ aValue * 1e3], AppLocale)
    else if aValue < 10
    then Result := Format( '%.2f s' , [ aValue      ], AppLocale)
    else if aValue < 60
    then Result := Format( '%.1f s' , [ aValue      ], AppLocale)
    else if aValue < 3600
    then Result := FormatDateTime( 'nn:ss'  , aValue / ( 24 * 3600), AppLocale)
    else Result := FormatDateTime( 'h:nn:ss', aValue / ( 24 * 3600), AppLocale);
  end;

  function  FreqToStr( const aValue: TSignal): string;
  begin
    if aValue = 0
    then Result := '0 Hz'
    else if Abs( aValue) < 0.1
    then Result := TimeToStr( 1 / aValue)
    else if Abs( aValue) < 1
    then Result := Format( '%.3f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 10
    then Result := Format( '%.3f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 100
    then Result := Format( '%.2f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 1000
    then Result := Format( '%.1f Hz', [ aValue], AppLocale)
    else if Abs( aValue) < 10000
    then Result := Format( '%.2f kHz', [ aValue / 1000], AppLocale)
    else Result := Format( '%.1f kHz', [ aValue / 1000], AppLocale);
  end;

  function  FreqToBPMStr( const aValue: TSignal): string;
  begin
    Result := Format( '%.0f', [ aValue * 60], AppLocale);
  end;

  function  ScaleIndexToData( aScale: Integer): TScale;
  begin
    Result := TBaseNotes;
    case aScale of
      scale_chromatic          : Result := [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11];
      scale_octatonic          : Result := [ 0, 1, 3, 4, 6, 7, 9, 10];
      scale_Istrian            : Result := [ 0, 1, 3, 4, 6, 7];
      scale_Altered            : Result := [ 0, 1, 3, 4, 6, 8, 10];
      scale_Locrian            : Result := [ 0, 1, 3, 5, 6, 8, 10];
      scale_Phrygian           : Result := [ 0, 1, 3, 5, 7, 8, 10];
      scale_Neapolitan_minor   : Result := [ 0, 1, 3, 5, 7, 8, 11];
      scale_Neapolitan_major   : Result := [ 0, 1, 3, 5, 7, 9, 11];
      scale_Persian            : Result := [ 0, 1, 4, 5, 6, 8, 11];
      scale_gypsy              : Result := [ 0, 1, 4, 5, 6, 9, 10];
      scale_Phrygian_dominant  : Result := [ 0, 1, 4, 5, 7, 8, 10];
      scale_Flamenco           : Result := [ 0, 1, 4, 5, 7, 8, 11];
      scale_Tritone            : Result := [ 0, 1, 4, 6, 7, 10];
      scale_Enigmatic          : Result := [ 0, 1, 4, 6, 8, 10, 11];
      scale_Iwato              : Result := [ 0, 1, 5, 6, 10];
      scale_Insen              : Result := [ 0, 1, 5, 7, 10];
      scale_blues_major        : Result := [ 0, 2, 3, 4, 7, 9];
      scale_Octatonic_2        : Result := [ 0, 2, 3, 5, 6, 8, 9, 10];
      scale_Half_diminished    : Result := [ 0, 2, 3, 5, 6, 8, 10];
      scale_minor              : Result := [ 0, 2, 3, 5, 7, 8, 10];
      scale_Harmonic_minor_1   : Result := [ 0, 2, 3, 5, 7, 8];
      scale_Dorian             : Result := [ 0, 2, 3, 5, 7, 9, 10];
      scale_Melodic_minor      : Result := [ 0, 2, 3, 5, 7, 10];
      scale_harmonic_minor_2   : Result := [ 0, 2, 3, 6, 7, 8, 10];
      scale_Hungarian_gyps     : Result := [ 0, 2, 3, 6, 7, 8, 11];
      scale_Algerian           : Result := [ 0, 2, 3, 6, 7, 8];
      scale_Ukrainian_Dorian   : Result := [ 0, 2, 3, 6, 7, 9, 10];
      scale_hungarian_gypsy    : Result := [ 0, 2, 3, 6, 7];
      scale_hung_gyp_augmented : Result := [ 0, 2, 3, 6, 8];
      scale_Hirajoshi          : Result := [ 0, 2, 3, 7, 8];
      scale_Major_Locrian      : Result := [ 0, 2, 4, 5, 6, 8, 10];
      scale_Major_bebop        : Result := [ 0, 2, 4, 5, 7, 8, 9, 11];
      scale_Adonai_malakh      : Result := [ 0, 2, 4, 5, 7, 8, 10];
      scale_Harmonic_major     : Result := [ 0, 2, 4, 5, 7, 8, 11];
      scale_Bebop_dominant     : Result := [ 0, 2, 4, 5, 7, 9, 10, 11];
      scale_Mixolydian         : Result := [ 0, 2, 4, 5, 7, 9, 10];
      scale_major              : Result := [ 0, 2, 4, 5, 7, 9, 11];
      scale_Acoustic           : Result := [ 0, 2, 4, 6, 7, 9, 10];
      scale_Lydian             : Result := [ 0, 2, 4, 6, 7, 9, 11];
      scale_Lydian_augmented   : Result := [ 0, 2, 4, 6, 8, 9, 11];
      scale_whole_tone         : Result := [ 0, 2, 4, 6, 8, 10];
      scale_Prometheus         : Result := [ 0, 2, 4, 6, 9, 10];
      scale_pentatonic_Major   : Result := [ 0, 2, 5, 7, 9];
      scale_Augmented          : Result := [ 0, 3, 4, 7, 8, 11];
      scale_blues_minor        : Result := [ 0, 3, 5, 6, 7, 10];
      scale_pentatonic_minor   : Result := [ 0, 3, 5, 7, 10];
      scale_Yo                 : Result := [ 0, 3, 5, 7, 11];
      scale_japanese           : Result := [ 0, 4, 5, 9, 11];
      else                       Result := [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11];
    end;
  end;

  function  ScaleIndexToName( aScale: Integer): string;
  begin
    Result := 'no_scale';
    case aScale of
      scale_chromatic          : Result := 'chromatic';
      scale_octatonic          : Result := 'octatonic';
      scale_Istrian            : Result := 'Istrian';
      scale_Altered            : Result := 'Altered';
      scale_Locrian            : Result := 'Locrian';
      scale_Phrygian           : Result := 'Phrygian';
      scale_Neapolitan_minor   : Result := 'Neapolitan_minor';
      scale_Neapolitan_major   : Result := 'Neapolitan_major';
      scale_Persian            : Result := 'Persian';
      scale_gypsy              : Result := 'gypsy';
      scale_Phrygian_dominant  : Result := 'Phrygian_dominant';
      scale_Flamenco           : Result := 'Flamenco';
      scale_Tritone            : Result := 'Tritone';
      scale_Enigmatic          : Result := 'Enigmatic';
      scale_Iwato              : Result := 'Iwato';
      scale_Insen              : Result := 'Insen';
      scale_blues_major        : Result := 'blues_major';
      scale_Octatonic_2        : Result := 'Octatonic_2';
      scale_Half_diminished    : Result := 'Half_diminished';
      scale_minor              : Result := 'minor';
      scale_Harmonic_minor_1   : Result := 'Harmonic_minor_1';
      scale_Dorian             : Result := 'Dorian';
      scale_Melodic_minor      : Result := 'Melodic_minor';
      scale_harmonic_minor_2   : Result := 'harmonic_minor_2';
      scale_Hungarian_gyps     : Result := 'Hungarian_gyps';
      scale_Algerian           : Result := 'Algerian';
      scale_Ukrainian_Dorian   : Result := 'Ukrainian_Dorian';
      scale_hungarian_gypsy    : Result := 'hungarian_gypsy';
      scale_hung_gyp_augmented : Result := 'hung_gyp_augmented';
      scale_Hirajoshi          : Result := 'Hirajoshi';
      scale_Major_Locrian      : Result := 'Major_Locrian';
      scale_Major_bebop        : Result := 'Major_bebop';
      scale_Adonai_malakh      : Result := 'Adonai_malakh';
      scale_Harmonic_major     : Result := 'Harmonic_major';
      scale_Bebop_dominant     : Result := 'Bebop_dominant';
      scale_Mixolydian         : Result := 'Mixolydian';
      scale_major              : Result := 'major';
      scale_Acoustic           : Result := 'Acoustic';
      scale_Lydian             : Result := 'Lydian';
      scale_Lydian_augmented   : Result := 'Lydian_augmented';
      scale_whole_tone         : Result := 'whole_tone';
      scale_Prometheus         : Result := 'Prometheus';
      scale_pentatonic_Major   : Result := 'pentatonic_Major';
      scale_Augmented          : Result := 'Augmented';
      scale_blues_minor        : Result := 'blues_minor';
      scale_pentatonic_minor   : Result := 'pentatonic_minor';
      scale_Yo                 : Result := 'Yo';
      scale_japanese           : Result := 'japanese';
    end;
  end;

  function  MakeSignalPair( aLeft, aRight: TSignal): TSignalPair; inline;
  begin
    Result.Left  := aLeft;
    Result.Right := aRight;
  end;

  function  MakeComplexSignal( aReal, anImag: TSignal): TComplex; inline;
  begin
    Result.Real := aReal;
    Result.Imag := anImag;
  end;

  procedure RegisterConverters( const aClasses: array of TKnobsConverterClass);
  begin
    LocConverters.RegisterConverters( aClasses);
  end;

  function  Converters: TKnobsConverters;
  begin
    Result := LocConverters;
  end;



const

  LutRange = 8192;          // [0,1] TSignal range maps to [0,LutRange] table section
  LutSize  = 2 * LutRange;  // [0,2] TSignal range maps to [0,LutSize ] table section

type

  TLut = array[ - LutSize .. LutSize + 1] of TSignal;

var

  UnitsToFrequencyTable           : TLut;
  UnitsToFrequencyLinTable        : TLut;
  UnitsToFrequencyCubedTable      : TLut;
  UnitsToPhaseDeltaTable          : TLut;
  UnitsToPhaseDeltaLinTable       : TLut;
  UnitsToSvfFreqTable             : TLut;
  UnitsToLfoSlowFrequencyTable    : TLut;
  UnitsToLfoSlowPhaseDeltaTable   : TLut;
  UnitsToLfoMediumFrequencyTable  : TLut;
  UnitsToLfoMediumPhaseDeltaTable : TLut;
  UnitsToLfoFastFrequencyTable    : TLut;
  UnitsToLfoFastPhaseDeltaTable   : TLut;
  UnitsToLfoBPMFrequencyTable     : TLut;
  UnitsToLfoBPMPhaseDeltaTable    : TLut;
  UnitsToLfoLinFrequencyTable     : TLut;
  UnitsToLfoLinPhaseDeltaTable    : TLut;
  UnitsToSineTable                : TLut;
  UnitsToCosineTable              : TLut;
  UnitsToScaledEnvTimeTable       : TLut;


  procedure ComputeTable( var aTable: TLut; const aFunc: TSignalFunction);
  var
    i : Integer;
  begin
    for i := Low( aTable) to High( aTable)
    do aTable[ i] := Normalize( aFunc( i / LutRange));
  end;

  function TableLookup( const aValue: TSignal; const aTable : TLut): TSignal; inline;
  var
    aLow   : Integer;
    aVal   : TSignal;
    aDelta : TSignal;
  begin
    aVal   := Clip( aValue * LutRange, Low( aTable), High( aTable) - 1);
    aLow   := Trunc( aVal);
    aDelta := aVal - aLow;
    Result := aTable[ aLow];
    Result := Result + aDelta * ( aTable[ aLow + 1] - Result);
  end;

  procedure ComputeTables;
  begin
    ComputeTable( UnitsToFrequencyTable          , UnitsToFrequency          );
    ComputeTable( UnitsToFrequencyLinTable       , UnitsToFrequencyLin       );
    ComputeTable( UnitsToFrequencyCubedTable     , UnitsToFrequencyCubed     );
    ComputeTable( UnitsToPhaseDeltaTable         , UnitsToPhaseDelta         );
    ComputeTable( UnitsToPhaseDeltaLinTable      , UnitsToPhaseDeltaLin      );
    ComputeTable( UnitsToSvfFreqTable            , UnitsToSvfFrequency       );
    ComputeTable( UnitsToLfoSlowFrequencyTable   , UnitsToLfoSlowFrequency   );
    ComputeTable( UnitsToLfoSlowPhaseDeltaTable  , UnitsToLfoSlowPhaseDelta  );
    ComputeTable( UnitsToLfoMediumFrequencyTable , UnitsToLfoMediumFrequency );
    ComputeTable( UnitsToLfoMediumPhaseDeltaTable, UnitsToLfoMediumPhaseDelta);
    ComputeTable( UnitsToLfoFastFrequencyTable   , UnitsToLfoFastFrequency   );
    ComputeTable( UnitsToLfoFastPhaseDeltaTable  , UnitsToLfoFastPhaseDelta  );
    ComputeTable( UnitsToLfoBPMFrequencyTable    , UnitsToLfoBPMFrequency    );
    ComputeTable( UnitsToLfoBPMPhaseDeltaTable   , UnitsToLfoBPMPhaseDelta   );
    ComputeTable( UnitsToLfoLinFrequencyTable    , UnitsToLfoLinFrequency    );
    ComputeTable( UnitsToLfoLinPhaseDeltaTable   , UnitsToLfoLinPhaseDelta   );
    ComputeTable( UnitsToSineTable               , UnitsToSine               );
    ComputeTable( UnitsToCosineTable             , UnitsToCosine             );
    ComputeTable( UnitsToScaledEnvTimeTable      , UnitsToScaledEnvTime      );
  end;

  function  LookupFrequency          ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToFrequencyTable          ); end;
  function  LookupFrequencyLin       ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToFrequencyLinTable       ); end;
  function  LookupFrequencyCubed     ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToFrequencyCubedTable     ); end;
  function  LookupPhaseDelta         ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToPhaseDeltaTable         ); end;
  function  LookupPhaseDeltaLin      ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToPhaseDeltaLinTable      ); end;
  function  LookupSvfFrequency       ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToSvfFreqTable            ); end;
  function  LookupLfoSlowFrequency   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoSlowFrequencyTable   ); end;
  function  LookupLfoSlowPhaseDelta  ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoSlowPhaseDeltaTable  ); end;
  function  LookupLfoMediumFrequency ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoMediumFrequencyTable ); end;
  function  LookupLfoMediumPhaseDelta( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoMediumPhaseDeltaTable); end;
  function  LookupLfoFastFrequency   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoFastFrequencyTable   ); end;
  function  LookupLfoFastPhaseDelta  ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoFastPhaseDeltaTable  ); end;
  function  LookupLfoBPMFrequency    ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoBPMFrequencyTable    ); end;
  function  LookupLfoBPMPhaseDelta   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoBPMPhaseDeltaTable   ); end;
  function  LookupLfoLinFrequency    ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoLinFrequencyTable    ); end;
  function  LookupLfoLinPhaseDelta   ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToLfoLinPhaseDeltaTable   ); end;
  function  LookupSine               ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToSineTable               ); end;
  function  LookupCosine             ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToCosineTable             ); end;
  function  LookupScaledEnvTime      ( const aValue: TSignal): TSignal; begin Result := TableLookup( aValue, UnitsToScaledEnvTimeTable      ); end;


  function  LookupEnvTime( aRange: TEnvRange; aValue: TSignal): TSignal;
  var
    p         : TSignal;
    anOutLow  : TSignal;
    anOutHigh : TSignal;
  begin
    case aRange of
      erSlow   : begin anOutLow := MinEnvTimeSlow  ; anOutHigh := MaxEnvTimeSlow  ; end;
      erMedium : begin anOutLow := MinEnvTimeMedium; anOutHigh := MaxEnvTimeMedium; end;
      else       begin anOutLow := MinEnvTimeFast  ; anOutHigh := MaxEnvTimeFast  ; end;
    end;
    p      := LookupScaledEnvTime( Clip( aValue, 0, 1));
    Result := anOutLow + p * ( anOutHigh - anOutLow);
  end;


  function  MidiFilterToStr( const aValue: TMidiFilter): string;
  var
    i : TMidiSeptet;
  begin
    Result := '{ ';
    for i := Low( TMidiSeptet) to High( TMidiSeptet)
    do begin
      if i in aValue
      then Result := Result + Format( '%d ', [ i], AppLocale);
    end;
    Result := Result + '}';
  end;

  procedure SetSystemRate( aValue: Integer);
  begin
    System_Rate        := aValue;                           // samples / s
    System_Rate_2      := System_Rate * 2;                  // samples / s
    Control_Decimation := 8;
    Control_Rate       := System_Rate / Control_Decimation; // samples / s
    ComputeTables;
  end;

  function  KnobRangeLin( const aPos, aStepCount: Integer; anOutLow, anOutHigh: TSignal): TSignal; inline;
  begin
    if aStepCount <= 0
    then Result := anOutLow
    else Result := anOutLow + ( aPos / aStepCount) * ( anOutHigh - anOutLow);
  end;

  function  KnobRangeExp( const aPos, aStepCount: Integer; anOutLow, anOutHigh: TSignal): TSignal; inline;
  var
    p : TSignal;
  begin
    if aStepCount <= 0
    then Result := anOutLow
    else begin
      p      := LookupScaledEnvTime( aPos / aStepCount);
      Result := anOutLow + p * ( anOutHigh - anOutLow);
    end;
  end;


type

  TConvUndefined = class( TKnobsNullConverter)
  public
    constructor Create;                                               override;
    function    PosToHint   ( aPos, aSteps: Integer): string;         override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TKnobsBaseConverter = class( TKnobsNullConverter)
  public
    function    PosToHint   ( aPos, aSteps: Integer): string;         override;
  end;

  TConvLinear0_1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLinear0_4 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLambda = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLinearMin1_1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLinearMin4_4 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvAdd0 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvAdd1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvOdd = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvQSteps = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvdB0_1 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvdB0_4 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvdB0_16 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvdB0_32 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvdB0_64 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvdBKS = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvdBResonator = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvOscFreqCoarse = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvOscLinCoarse = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvOscCubedCoarse = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseFast = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseMedium = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseSlow = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseBPM = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvLfoFreqCoarseLin = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvNoiseColor = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvEnvTimeFast = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvEnvTimeMedium = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvEnvTimeSlow = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvMidiChIn = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvMidiChOut = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvNoteName = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvNoteName12 = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvDelayTime = class( TKnobsBaseConverter)
  public
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TConvDelayTimeShort = class( TConvDelayTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvDelayTimeMedium = class( TConvDelayTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvDelayTimeLong = class( TConvDelayTime)
  public
    constructor Create;                                               override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvFilterFreq = class( TConvOscFreqCoarse)
  public
    constructor Create;                                               override;
  end;

  TConvAvgFilterAlpha = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvCents = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvSimple = class( TKnobsBaseConverter)
  public
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TConvMute = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TConvSolo = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TConvEnvMute = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TConvOutType = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TConvEnvOutType = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TConvWaveShape = class( TConvSimple)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
  end;

  TTalkieFrameRate = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;

  TTalkieSampleRate = class( TKnobsBaseConverter)
  public
    constructor Create;                                               override;
    function    PosToDisplay( aPos, aSteps: Integer): string;         override;
    function    PosToValue  ( aPos, aSteps: Integer): TSignal;        override;
  end;


{ ========
  TKnobsNullConverter = class
  private
    FSearched : string;
  protected
    FName     : string;
    FRate     : TSignalRate;
  public
    property    Name     : string  read FName;
    property    Rate     : TSignal read GetRate;
    property    Searched : string  read FSearched write FSearched;
  private
}

    function    TKnobsNullConverter.GetRate: TSignal;
    begin
      case FRate of
        srControl : Result := Control_Rate;
        else        Result := System_Rate;
      end;
    end;

//  public

    constructor TKnobsNullConverter.Create; // virtual;
    begin
      FName     := '';
      FSearched := '';
      FRate     := srAudio;
    end;

    function    TKnobsNullConverter.PosToHint( aPos, aSteps: Integer): string; // virtual;
    begin
      Result := IntToStr( aPos);
    end;

    function    TKnobsNullConverter.PosToDisplay( aPos, aSteps: Integer): string; // virtual;
    begin
      Result := IntToStr( aPos);
    end;

    function    TKnobsNullConverter.PosToValue( aPos, aSteps: Integer): TSignal; // virtual;
    begin
      Result := aPos;
    end;


{ ========
  TKnobsConverterUndefined = class( TKnobsNullConverter)
  public
}
    constructor TConvUndefined.Create; // override;
    begin
      FName := 'undefined converter';
      FRate := srAudio;
    end;

    function    TConvUndefined.PosToHint( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( 'undefined %s'^M'does %s'^M'%f', [ Searched, inherited, PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvUndefined.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( 'undefined %s does %s [%f]', [ Searched, inherited, PosToValue( aPos, aSteps)], AppLocale);
    end;


{ ========
  TKnobsBaseConverter = class( TKnobsNullConverter)
  public
}

    function    TKnobsBaseConverter.PosToHint( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( ' value: %s '^M''^M' position: %d ', [ PosToDisplay( aPos, aSteps), aPos], AppLocale);
    end;


{ ========
  TConvLinear0_1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinear0_1.Create; // override;
    begin
      FName := 'Linear[0,1]';
      FRate := srAudio;
    end;

    function    TConvLinear0_1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvLinear0_1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TConvLinear0_4 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinear0_4.Create; // override;
    begin
      FName := 'Linear[0,4]';
      FRate := srAudio;
    end;

    function    TConvLinear0_4.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvLinear0_4.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 4);
    end;


{ ========
  TConvLambda = class( TKnobsBaseConverter)
  public
}

    constructor TConvLambda.Create; // override;
    begin
      FName := 'Lambda';
      FRate := srAudio;
    end;

    function    TConvLambda.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvLambda.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 5, 1);
    end;


{ ========
  TConvLineMin1_1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinearMin1_1.Create; // override;
    begin
      FName := 'Linear[-1,1]';
      FRate := srAudio;
    end;

    function    TConvLinearMin1_1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.3f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvLinearMin1_1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -1, 1);
    end;



{ ========
  TConvLineMin4_4 = class( TKnobsBaseConverter)
  public
}

    constructor TConvLinearMin4_4.Create; // override;
    begin
      FName := 'Linear[-4,4]';
      FRate := srAudio;
    end;

    function    TConvLinearMin4_4.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvLinearMin4_4.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -4, 4);
    end;



{ ========
  TConvAdd0 = class( TKnobsBaseConverter)
  public
}

    constructor TConvAdd0.Create; // override;
    begin
      FName := 'Add0';
      FRate := srAudio;
    end;

    function    TConvAdd0.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.0f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvAdd0.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos;
    end;


{ ========
  TConvAdd1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvAdd1.Create; // override;
    begin
      FName := 'Add1';
      FRate := srAudio;
    end;

    function    TConvAdd1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.0f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvAdd1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos + 1;
    end;


{ ========
  TConvOdd = class( TKnobsBaseConverter)
  public
}

    constructor TConvOdd.Create; // override;
    begin
      FName := 'Odd';
      FRate := srAudio;
    end;

    function    TConvOdd.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := IntToStr( 2 * aPos + 1);
    end;

    function    TConvOdd.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := 2 * aPos + 1;
    end;


{ ========
  TConvQSteps = class( TKnobsBaseConverter)
  public
}

    constructor TConvQSteps.Create; // override;
    begin
      FName := 'QSteps';
      FRate := srAudio;
    end;

    function    TConvQSteps.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      R : TRational;
    begin
      R := PosToValue( aPos, aSteps) / 64;
      R.Invert;
      Result := Format( '%s', [ R.AsString], AppLocale);
    end;

    function    TConvQSteps.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos + 1;
    end;


{ ========
  TConvdB0_1 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_1.Create; // override;
    begin
      FName := 'dB[0,1]';
      FRate := srAudio;
    end;

    function    TConvdB0_1.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := SignalTodB( aValue);
      Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;

    function    TConvdB0_1.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TConvdB0_4 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_4.Create; // override;
    begin
      FName := 'dB[0,4]';
      FRate := srAudio;
    end;

    function    TConvdB0_4.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := SignalTodB( aValue);
      Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;

    function    TConvdB0_4.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 4);
    end;


{ ========
  TConvdB0_16 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_16.Create; // override;
    begin
      FName := 'dB[0,16]';
      FRate := srAudio;
    end;

    function    TConvdB0_16.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := SignalTodB( aValue);
      Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;

    function    TConvdB0_16.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 16);
    end;


{ ========
  TConvdB0_32 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_32.Create; // override;
    begin
      FName := 'dB[0,32]';
      FRate := srAudio;
    end;

    function    TConvdB0_32.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := SignalTodB( aValue);
      Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;

    function    TConvdB0_32.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 32);
    end;

{ ========
  TConvdB0_64 = class( TKnobsBaseConverter)
  public
}

    constructor TConvdB0_64.Create; // override;
    begin
      FName := 'dB[0,64]';
      FRate := srAudio;
    end;

    function    TConvdB0_64.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := SignalTodB( aValue);
      Result := Format( '%.1f dB', [ aValue], AppLocale);
    end;

    function    TConvdB0_64.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 1600);
    end;

{ ========
  TConvdBKS = class( TKnobsBaseConverter)
  public
}

    constructor TConvdBKS.Create; // override;
    begin
      FName := 'dB KS';
      FRate := srAudio;
    end;

    function    TConvdBKS.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := SignalTodB( aValue);
      Result := Format( '%.2f dB', [ aValue], AppLocale);
    end;

    function    TConvdBKS.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0.8, 1);
    end;



{ ========
  TConvdbResonator = class( TKnobsBaseConverter)
  public
}

    constructor TConvdbResonator.Create; // override;
    begin
      FName := 'dB Resonator';
      FRate := srAudio;
    end;

    function    TConvdbResonator.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := SignalTodB( aValue);
      Result := Format( '%.6f dB', [ aValue], AppLocale);
    end;

    function    TConvdbResonator.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0.999, 0.999999999999999);
    end;



{ ========
  TConvOscFreqCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvOscFreqCoarse.Create; // override;
    begin
      FName := 'OscFreqCoarse';
      FRate := srAudio;
    end;

    function    TConvOscFreqCoarse.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvOscFreqCoarse.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvOscLinCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvOscLinCoarse.Create; // override;
    begin
      FName := 'OscLinCoarse';
      FRate := srAudio;
    end;

    function    TConvOscLinCoarse.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupFrequencyLin( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvOscLinCoarse.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvOscCubedCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvOscCubedCoarse.Create; // override;
    begin
      FName := 'OscCubedCoarse';
      FRate := srAudio;
    end;

    function    TConvOscCubedCoarse.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupFrequencyCubed( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvOscCubedCoarse.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvLfoFreqCoarse = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseFast.Create; // override;
    begin
      FName := 'LfoFreqCoarseFast';
      FRate := srControl;
    end;

    function    TConvLfoFreqCoarseFast.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoFastFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvLfoFreqCoarseFast.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvLfoFreqCoarseMedium = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseMedium.Create; // override;
    begin
      FName := 'LfoFreqCoarseMedium';
      FRate := srControl;
    end;

    function    TConvLfoFreqCoarseMedium.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoMediumFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvLfoFreqCoarseMedium.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvLfoFreqCoarseSlow = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseSlow.Create; // override;
    begin
      FName := 'LfoFreqCoarseSlow';
      FRate := srControl;
    end;

    function    TConvLfoFreqCoarseSlow.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoSlowFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvLfoFreqCoarseSlow.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;



{ ========
  TConvLfoFreqCoarseBPM = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseBPM.Create; // override;
    begin
      FName := 'LfoFreqCoarseBPM';
      FRate := srControl;
    end;

    function    TConvLfoFreqCoarseBPM.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoBPMFrequency( aValue);
      Result := FreqToBPMStr( aValue);
    end;

    function    TConvLfoFreqCoarseBPM.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;



{ ========
  TConvLfoFreqCoarseLin = class( TKnobsBaseConverter)
  public
}

    constructor TConvLfoFreqCoarseLin.Create; // override;
    begin
      FName := 'LfoFreqCoarseLin';
      FRate := srControl;
    end;

    function    TConvLfoFreqCoarseLin.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      aValue := PosToValue( aPos, aSteps);
      aValue := LookupLfoLinFrequency( aValue);
      Result := FreqToStr( aValue);
    end;

    function    TConvLfoFreqCoarseLin.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;



{ ========
  TConvNoiseColor = class( TKnobsBaseConverter)
  public
}

    constructor TConvNoiseColor.Create; // override;
    begin
      FName := 'NoiseColor';
      FRate := srAudio;
    end;

    function    TConvNoiseColor.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      if aPos = 0
      then Result := 'off'
      else if aPos = aSteps
      then Result := 'white'
      else Result := Format( '%.2f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvNoiseColor.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TConvEnvTimeFast = class( TKnobsBaseConverter)
  public
}

    constructor TConvEnvTimeFast.Create; // override;
    begin
      FName := 'EnvTimeFast';
      FRate := srAudio;
    end;

    function    TConvEnvTimeFast.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      if aSteps <= 1
      then aValue := MinEnvTimeFast
      else aValue := LookupEnvTime( erFast, aPos / ( aSteps - 1));
      Result := TimeToStr( aValue);
    end;

    function    TConvEnvTimeFast.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TConvEnvTimeMedium = class( TKnobsBaseConverter)
  public
}

    constructor TConvEnvTimeMedium.Create; // override;
    begin
      FName := 'EnvTimeMedium';
      FRate := srAudio;
    end;

    function    TConvEnvTimeMedium.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      if aSteps <= 1
      then aValue := MinEnvTimeMedium
      else aValue := LookupEnvTime( erMedium, aPos / ( aSteps - 1));
      Result := TimeToStr( aValue);
    end;

    function    TConvEnvTimeMedium.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TConvEnvTimeSlow = class( TKnobsBaseConverter)
  public
}

    constructor TConvEnvTimeSlow.Create; // override;
    begin
      FName := 'EnvTimeSlow';
      FRate := srAudio;
    end;

    function    TConvEnvTimeSlow.PosToDisplay( aPos, aSteps: Integer): string; // override;
    var
      aValue : TSignal;
    begin
      if aSteps <= 1
      then aValue := MinEnvTimeSlow
      else aValue := LookupEnvTime( erSlow, aPos / ( aSteps - 1));
      Result := TimeToStr( aValue);
    end;

    function    TConvEnvTimeSlow.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TConvMidiCCIn = class( TKnobsBaseConverter)
  public
}

    constructor TConvMidiChIn.Create; // override;
    begin
      FName := 'MidiChIn';
      FRate := srAudio;
    end;

    function    TConvMidiChIn.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      if  aPos = aSteps - 1
      then Result := 'omni'
      else Result := Format( 'ch %.0f', [ PosToValue( aPos, aSteps) + 1], AppLocale);
    end;

    function    TConvMidiChIn.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos;
    end;


{ ========
  TConvMidiCCOut = class( TKnobsbaseConverter)
  public
}

    constructor TConvMidiChOut.Create; // override;
    begin
      FName := 'MidiChOut';
      FRate := srAudio;
    end;

    function    TConvMidiChOut.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( 'ch %.0f', [ PosToValue( aPos, aSteps) + 1], AppLocale);
    end;

    function    TConvMidiChOut.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos;
    end;



{ ========
  TConvNoteName = class( TKnobsBaseConverter)
  public
}

    constructor TConvNoteName.Create; // override;
    begin
      FName := 'NoteName';
      FRate := srAudio;
    end;

    function    TConvNoteName.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := UnitsToNoteName( PosToValue( aPos, aSteps));
    end;

    function    TConvNoteName.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvNoteName12 = class( TKnobsBaseConverter)
  public
}

    constructor TConvNoteName12.Create; // override;
    begin
      FName := 'NoteName12';
      FRate := srAudio;
    end;

    function    TConvNoteName12.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := UnitsToNoteName12( PosToValue( aPos, aSteps));
    end;

    function    TConvNoteName12.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps, 0, 1);
    end;


{ ========
  TConvDelayTime = class( TKnobsBaseConverter)
  public
}

    function    TConvDelayTime.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := TimeToStr( PosToValue( aPos, aSteps));
    end;


{ ========
  TConvDelayTimeShort = class( TKnobsBaseConverter)
  public
}

    constructor TConvDelayTimeShort.Create; // override;
    begin
      FName := 'DelayTimeShort';
      FRate := srAudio;
    end;

    function    TConvDelayTimeShort.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 1e-4, 3e-2);
    end;


{ ========
  TConvDelayTimeMedium = class( TKnobsBaseConverter)
  public
}

    constructor TConvDelayTimeMedium.Create; // override;
    begin
      FName := 'DelayTimeMedium';
      FRate := srAudio;
    end;

    function    TConvDelayTimeMedium.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 1e-2, 3);
    end;


{ ========
  TConvDelayTimeLong = class( TKnobsBaseConverter)
  public
}

    constructor TConvDelayTimeLong.Create; // override;
    begin
      FName := 'DelayTimeLong';
      FRate := srAudio;
    end;

    function    TConvDelayTimeLong.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 1, 3e2);
    end;


{ ========
  TConvFilterFreq = class( TKnobsBaseConverter)
  public
}

    constructor TConvFilterFreq.Create; // override;
    begin
      FName := 'FilterFreq';
      FRate := srAudio;
    end;


{ ========
  TConvAvgFilterFreq = class( TKnobsBaseConverter)
  public
}

    constructor TConvAvgFilterAlpha.Create; // override;
    begin
      FName := 'FilterAvgAlpha';
      FRate := srAudio;
    end;

    function    TConvAvgFilterAlpha.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.5f', [ PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvAvgFilterAlpha.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeExp( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TConvCents = class( TKnobsBaseConverter)
  public
}

    constructor TConvCents.Create; // override;
    begin
      FName := 'Cents';
      FRate := srAudio;
    end;

    function    TConvCents.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.0f ct', [ 50 * 256 * PosToValue( aPos, aSteps)], AppLocale);
    end;

    function    TConvCents.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, -1 / 256, 1 / 256);
    end;


{ ========
  TConvSimple = class( TKnobsBaseConverter)
  public
}

    function    TConvSimple.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := aPos;
    end;


{ ========
  TConvMute = class( TKnobsBaseConverter)
  public
}

    constructor TConvMute.Create; // override;
    begin
      FName := 'mute';
      FRate := srControl;
    end;

    function    TConvMute.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'active';
        1 :  Result := 'muted';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvSolo = class( TKnobsBaseConverter)
  public
}

    constructor TConvSolo.Create; // override;
    begin
      FName := 'solo';
      FRate := srControl;
    end;

    function    TConvSolo.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'normal';
        1 :  Result := 'solo';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvEnvMute = class( TConvSimple)
  public
}

    constructor TConvEnvMute.Create; // override;
    begin
      FName := 'envmute';
      FRate := srControl;
    end;

    function    TConvEnvMute.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'active';
        1 :  Result := 'muted';
        2 :  Result := 'pass through';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvOutType = class( TConvSimple)
  public
}

    constructor TConvOutType.Create; // override;
    begin
      FName := 'outtype';
      FRate := srControl;
    end;

    function    TConvOutType.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'Normal';
        1 :  Result := 'Inverted';
        2 :  Result := 'Positive';
        3 :  Result := 'Positve Inverted';
        4 :  Result := 'Negative';
        5 :  Result := 'Negative Inverted';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvEnvOutType = class( TConvSimple)
  public
}

    constructor TConvEnvOutType.Create; // override;
    begin
      FName := 'envouttype';
      FRate := srControl;
    end;

    function    TConvEnvOutType.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'Positive';
        1 :  Result := 'Positve Inverted';
        2 :  Result := 'Negative';
        3 :  Result := 'Negative Inverted';
        else Result := 'unknown';
      end;
    end;


{ ========
  TConvWaveShape = class( TConvSimple)
  public
}

    constructor TConvWaveShape.Create; // override;
    begin
      FName := 'waveshape';
      FRate := srControl;
    end;

    function    TConvWaveShape.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      case aPos of
        0 :  Result := 'Sine';
        1 :  Result := 'Tri';
        2 :  Result := 'Saw';
        3 :  Result := 'Square';
        else Result := 'unknown';
      end;
    end;

{ ========
  TTalkieFrameRate = class( TKnobsBaseConverter)
  public
}

    constructor TTalkieFrameRate.Create; // override;
    begin
      FName := 'talkie frame rate';
      FRate := srControl;
    end;

    function    TTalkieFrameRate.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.2f f/s', [ 80.0 * PosToValue( aPos, aSteps)]);
    end;

    function    TTalkieFrameRate.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;


{ ========
  TTalkieSampleRate = class( TKnobsBaseConverter)
  public
}

    constructor TTalkieSampleRate.Create; // override;
    begin
      FName := 'talkie sample rate';
      FRate := srControl;
    end;

    function    TTalkieSampleRate.PosToDisplay( aPos, aSteps: Integer): string; // override;
    begin
      Result := Format( '%.1f s/s', [ 16000.0 * PosToValue( aPos, aSteps)]);
    end;

    function    TTalkieSampleRate.PosToValue( aPos, aSteps: Integer): TSignal; // override;
    begin
      Result := KnobRangeLin( aPos, aSteps - 1, 0, 1);
    end;



{ ========
  TKnobsConverters = class( TStringList)
  private
 }

    function    TKnobsConverters.Converter( const anIndex: string): TKnobsNullConverter;
    var
      p : Integer;
    begin
      p  := IndexOf( anIndex);
      if p < 0
      then begin
        p  := IndexOf( 'undefined converter');
        if p >= 0
        then (Objects[ p] as TKnobsNullConverter).Searched := anIndex
        else raise ENoDefaultConverter.Create( 'TKnobsConverters should at the very least have a default converter ("undefined converter", TKnobsNullConverter).');
      end;
      Result := Objects[ p] as TKnobsNullConverter;
    end;

 // public

    constructor TKnobsConverters.Create;
    begin
      inherited;
      OwnsObjects   := True;
      Duplicates    := dupError;
      Sorted        := True;
      CaseSensitive := False;
      RegisterConverters(
        [
          TKnobsNullConverter,      TConvUndefined,           TConvLinear0_1,           TConvLinear0_4,
          TConvLambda,              TConvLinearMin1_1,        TConvLinearMin4_4,        TConvdB0_1,
          TConvdB0_4,               TConvdB0_16,              TConvdB0_32,              TConvdB0_64,
          TConvdBKS,                TConvdBResonator,         TConvOscFreqCoarse,       TConvNoiseColor,
          TConvOscLinCoarse,        TConvOscCubedCoarse,      TConvLfoFreqCoarseFast,   TConvLfoFreqCoarseMedium,
          TConvLfoFreqCoarseSlow,   TConvLfoFreqCoarseBPM,    TConvLfoFreqCoarseLin,    TConvEnvTimeFast,
          TConvEnvTimeMedium,       TConvEnvTimeSlow,         TConvAdd0,                TConvAdd1,
          TConvOdd,                 TConvQSteps,              TConvMidiChIn,            TConvMidiChOut,
          TConvNoteName,            TConvNoteName12,          TConvDelayTimeShort,      TConvDelayTimeMedium,
          TConvDelayTimeLong,       TConvFilterFreq,          TConvAvgFilterAlpha,      TConvCents,
          TConvMute,                TConvSolo,                TConvEnvMute,             TConvOutType,
          TConvEnvOutType,          TConvWaveShape,           TTalkieFrameRate,         TTalkieSampleRate
        ]
      );
    end;

    procedure   TKnobsConverters.RegisterConverter( aClass: TKnobsConverterClass);
    var
      C: TKnobsNullConverter;
    begin
      C := aClass.Create;
      AddObject( C.Name, C);
    end;

    procedure   TKnobsConverters.RegisterConverters( const aClasses: array of TKnobsConverterClass);
    var
      C: TKnobsConverterClass;
    begin
      for C in aClasses
      do RegisterConverter( C);
    end;

    function    TKnobsConverters.PosToHint( const aName: string; aPos, aSteps: Integer): string;
    begin
      Result := Converter( aName).PosToHint( aPos, aSteps);
    end;

    function    TKnobsConverters.PosToDisplay( const aName: string; aPos, aSteps: Integer): string;
    begin
      Result := Converter( aName).PosToDisplay( aPos, aSteps);
    end;

    function    TKnobsConverters.PosToValue( const aName: string; aPos, aSteps: Integer): TSignal;
    begin
      Result := Converter( aName).PosToValue( aPos, aSteps);
    end;

    function    TKnobsConverters.Rate( const aName: string): TSignal;
    begin
      Result := Converter( aName).Rate;
    end;


  // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  procedure InitializeUnit;
  begin
    ReferenceA     := 440.0;
    NotesPerOctave := 12.0;
    MiddleNote     := 69;
    SetSystemRate( 44100);
  end;

  procedure CreateConverters;
  begin
    LocConverters := TKnobsConverters.Create;
  end;

  procedure FreeConverters;
  begin
    FreeAndNil( LocConverters);
  end;


initialization

  CreateConverters;
  InitializeUnit;

  // TestBinary;

finalization

  FreeConverters;

end.

